package WWW::GameFAQs::Archive::MessageList;

=head1 COPYRIGHT

Copyright 2010 Kodi Arfer.

This program is free software. You may copy or redistribute it under the same terms as Perl itself.

=cut

use List::Util 'sum';
use URI::Escape 'uri_escape_utf8';
use DBIx::Simple;
use DateTime;
use WWW::GameFAQs::Archive::Internal qw(htmlsafe div time_zone);

use parent 'Exporter';
@WWW::GameFAQs::Archive::MessageList::EXPORT_OK =
    qw(message_list);

sub message_list
#
# ~~~Mandatory arguments
# database => filepath
# topic => int
#   This argument will be validated.
# topic_list => urlfmt
# options_page => url
# topics_per_page => int
#
# ~~~Optional arguments
# post_ids => fmt
#   No URL escaping needed.
#
# ~~~Returns
# (Topic title [HTML-escaped], HTML message list)
#
   {my %o = @_;
    my $time_zone = time_zone;
    my $pfmt = htmlsafe($o{post_ids} or 'post%03d');

    my @error = (undef, div 'wgfa-error wgfa-no-such-topic',
        '<span><strong>Error:</strong> No such topic.</span>');
    $o{topic} = $o{topic} =~ m!\A(\d+)\z!
      ? $1
      : return @error;

    # Read from the database.
    
    my ($topic_list_page, $title, $closed, $poll_question, @poll, @rows) = (0);
       {my $db = DBIx::Simple->connect("dbi:SQLite:dbname=$o{database}")
            or die DBIx::Simple->error;
        $db->{sqlite_unicode} = 1;
        # Get title, closure state, and poll question.
        my $query = 'select title, closed, pollquestion from Topics where topic = ?';
        ($title, $closed, $poll_question) = ($db->query($query, $o{topic})
                or die DBIx::Simple->error)->flat;
        # Get the poll.
        if (defined $poll_question)
           {$query = 'select option_text, votes from PollOptions where topic = ? order by option_n';
            @poll = ($db->query($query, $o{topic})
                or die DBIx::Simple->error)->hashes;}
        # Get the posts.
        $query = 'select post, username, time, message, edited
           from Posts, Users
           where author = userid
                 and topic = ?
           order by post';
        @rows = ($db->query($query, $o{topic})
                or die DBIx::Simple->error)->arrays;
        @rows or return @error;
        # Get the page of the topic list on which this topic
        # appears.
        $query = 'select Topics.topic
           from Topics, Posts
           where Topics.topic = Posts.topic
                 and post = (select max(post)
                     from Posts
                     where topic = Topics.topic)
           order by Posts.time desc';
        my $result = $db->query($query) || die DBIx::Simple->error;
        my $n;
        ++$n until $result->list == $o{topic};
        $topic_list_page = int $n / $o{topics_per_page};}

    # Construct the page.

    my $s = div 'wgfa-back-to-topic-list',
        sprintf '<a href="%s">Topic List</a>',
        htmlsafe(sprintf $o{topic_list}, $topic_list_page);

    if (@poll)
       {$s .= "<div class='wgfa-poll-results'>\n";
        $s .= sprintf "<div class='wgfa-poll-question'>%s %s</div>\n",
            '<span class="wgfa-poll-marker">Poll:</span>',
            htmlsafe $poll_question;
        $s .= "<table class='wgfa-poll-results'>\n";
        foreach (@poll)
           {$s .= sprintf '<tr><td class="wgfa-votecount">%s<td>%s',
                $_->{votes} || '',
                htmlsafe $_->{option_text};}
        $s .= sprintf '<tr class="wgfa-votetotal"><td class="wgfa-votecount">%s<td>[Total]',
            sum map {$_->{votes}} @poll;
        $s .= "</table></div>\n";}

    $s .= "<div class='wgfa-message-list'>\n";
    foreach (@rows)
       {my ($post, $user, $time, $message, $edited) = @$_;
        my $header_class = 'wgfa-message-header';
        $edited and $header_class .= ' wgfa-message-header-edited';
        $s .= div 'wgfa-message',
            div($header_class, join ' | ',
                sprintf("<a id='$pfmt' href='#%s'>#%03d</a>",
                    $post,
                    uri_escape_utf8(sprintf $pfmt, $post),
                    $post),
                "<span class='wgfa-username'>$user</span>",
                sprintf('<time class="wgfa-full-timestamp" datetime="%s">%s</time>',
                    DateTime->from_epoch(epoch => $time)->iso8601() . 'Z',
                    DateTime->
                        from_epoch
                           (epoch => $time,
                            time_zone => $time_zone)->
                        format_cldr('ccc d MMM y @ h:mm:ss a (z)')),
                ($edited ? ('(edited)') : ())),
            div('wgfa-message-body', $message);
        $s .= "\n\n";}

    $closed and $s .= div 'wgfa-notification wgfa-closed-notice',
        '<span>This topic was closed before it was purged.</span>';

    $s .= "</div>\n";
    $s .= div 'wgfa-options-link', sprintf '<a href="%s">%s</a>',
        htmlsafe($o{options_page}),
        'Options';

    return ($title, $s);}

1;
