package WWW::GameFAQs::Archive::Spider;

=head1 COPYRIGHT

Copyright 2010 Kodi Arfer.

This program is free software. You may copy or redistribute it under the same terms as Perl itself.

=cut

use strict;
use Encode qw(encode decode);
use Carp;
use HTTP::Cookies;
use HTML::Entities;
use DBIx::Simple;
use Time::y2038;
use XML::Twig;
use WWW::GameFAQs::Archive::Internal qw(div);

use parent 'Exporter';

@WWW::GameFAQs::Archive::Spider::EXPORT_OK =
    qw(read_gamefaqs_timestamp update_archive);

package WWW::GameFAQs::Archive::Spider::Mech;
    use parent 'WWW::Mechanize';
    $WWW::GameFAQs::Archive::Spider::Mech::max_delay = 3; # Seconds.
    sub get
    # Sleeps after getting.
       {my ($self, @args) = @_;
        my $v = $self->SUPER::get(@args);
        sleep rand $WWW::GameFAQs::Archive::Spider::Mech::max_delay;
        return $v;}
package WWW::GameFAQs::Archive::Spider;

sub note;
sub sql;

sub flatsql
   {sql(@_)->flat}

sub runsub (&)
   {return $_[0]->();}

sub read_gamefaqs_timestamp
# Parses two timestamp formats that GameFAQs uses.
# Takes the timestamp. (It's assumed that the user's timezone is set to UTC.)
# Returns seconds since the epoch.
   {my $str = shift;
    my ($month, $day, $year, $hour, $min, $sec, $ampm) =
        $str =~ m!(\d+)/(\d+)/(\d+) (\d+):(\d+):(\d+)&nbsp;([AP]M)!;
    unless ($month)
       {($month, $day, $hour, $min, $ampm) =
            $str =~ m!(\d+)/(\d+) (\d+):(\d+)([AP]M)!
          or croak "Unrecognized time: $str\n";
        my @g = gmtime;
        my $current_month = $g[4] + 1;
        my $current_year = $g[5] + 1900;
        $year = $current_year - ($current_month < $month ? 1 : 0);
        $sec = 59;}
          # We want to err on the side of thinking something is
          # newer than it really is.
    return timegm
        $sec, $min,
        ($ampm eq 'PM' ? 12 : 0) + ($hour == 12 ? 0 : $hour),
        $day, $month - 1, $year - 1900;}

sub fix_blockquotes; sub fix_blockquotes
   {my $x = shift()->first_child;
    do
       {if ($x->gi eq 'cite'
                and defined $x->next_sibling
                and $x->next_sibling->gi eq 'br'
                and defined $x->next_sibling->next_sibling
                and $x->next_sibling->next_sibling->gi eq 'blockquote')
           {my $container = new XML::Twig::Elt('blockquote',
               {class => 'wgfa-quote-headed'});
            $x->next_sibling->delete; # Just a <br>.
            my $attribution = $x->copy;
            my $bq = $x->next_sibling->cut;
            $x->replace_with($container);
            $x = $container;

            $attribution->set_gi('div');
            $attribution->set_att(class => 'wgfa-quote-head');

            fix_blockquotes $bq;
            $bq->set_gi('div');
            $bq->set_att(class => 'wgfa-quote-tail');

            $x->set_content($attribution, $bq);}
        elsif ($x->gi eq 'blockquote')
           {fix_blockquotes $x;
            $x->set_att(class => 'wgfa-quote-headless');
            # Wrap the children of $x in a <div> for markup
            # validity.
            $x->set_content(XML::Twig::Elt->new('div',
                $x->cut_children));}}
    while defined($x = $x->next_sibling);}

sub digest_post
# Munges the markup of the content of a GameFAQs post.
   {my $s = sprintf '<doc>%s</doc>', encode 'UTF-8', shift;
    my $r = XML::Twig->new->parse($s)->root;
    fix_blockquotes $r;
    $s = join '', map {$_->sprint} $r->children;
    $s =~ s!<br/>!<br>!g;
    $s;}

sub update_archive
#
# ~~~~Mandatory arguments
# database => filepath
# email_address => string
# password => string
# username => string
# posts_per_page => int
#   Should match the user's setting.
#
# ~~~~Optional arguments
# cookie_jar => filename
# log_to => filehandle
#
   {my %o = @_;
    my $loghandle = $o{log_to};
    local *note = $o{log_to}
      ? sub {print $loghandle @_, "\n";}
      : sub {};

    my $checktime = time;
    note 'Starting: ', scalar localtime $checktime;

    # Set up the database connection and fetch some metadata.
    # print "alive 1\n";
    my $db = DBIx::Simple->connect("dbi:SQLite:dbname=$o{database}")
        or croak DBIx::Simple->error;
    # print "alive 2\n";
    local *sql = sub {$db->query(@_) or croak(DBIx::Simple->error)};
    $db->{sqlite_unicode} = 1;
    sql 'pragma foreign_keys = on';
    my ($last_checked) = flatsql
        q(select mval from Misc where mkey = 'last_checked');
    my ($board) = flatsql
        q(select mval from Misc where mkey = 'board');
    $db->begin;
    
    # Create a user agent and log in if necessary
    my $mech = new WWW::GameFAQs::Archive::Spider::Mech
       (stackdepth => 0,
        autocheck => 1,
        $o{cookie_jar}
          ? (cookie_jar => new HTTP::Cookies
             (file => $o{cookie_jar},
              autosave => 1))
          : ());
    $mech->{cookie_jar}->set_cookie(0,
        'throwback', '1',
        '/', '.gamefaqs.com',
        undef, 1, 0, 1e9, '', {});
    $mech->get("http://www.gamefaqs.com/boards/$board");
    my $topic_list = $mech->content;
    unless ($topic_list =~ /Welcome, \Q$o{username}\E/)
       {note 'Logging in.';
        $mech->submit_form(with_fields =>
           {EMAILADDR => $o{email_address},
            PASSWORD => $o{password}});
        $topic_list = $mech->content;
        $topic_list =~ /Welcome, \Q$o{username}\E/ or croak "Login phailed.\n";}
    
    for (;;)
      # Loop through the pages of the board.
       {my $next_board_page =
            $mech->find_link(text => 'Next Page') ||
            $mech->find_link(text => 'Last Page');
        while ($topic_list =~ m!
                /default/topic (_poll)? (_closed)? \.gif " \s alt=""/>
                </td> <td [^>]* >
                <a \s (?: style = " [^"]* " \s)?
                href=" (/boards/$board/(\d+)) ">
                (.+?)
                </a>
                .+? td \s class="lastpost">
                (.+?)
                </td> !gx)
          # Loop through the topics listed on this page.
           {# print "alive 3\n";
            my ($is_poll, $closed, $topic_url, $topic, $title) =
               (!!$1, !!$2, decode_entities($3), $4, $5);
            my $time = read_gamefaqs_timestamp $6;
            # print "alive 3.1\n";
            sql 'insert or ignore into Topics values (??)',
                $topic, $title, undef, 0;
            # print "alive 3.2\n";
            $closed and sql
                'update Topics set closed = 1 where topic = ?',
                $topic;
            if ($is_poll)
              # This has to be done separately from post-scraping
              # because we can't use the last-post timestamp
              # to figure out if we have to visit the topic for
              # new votes.
               {$mech->get("$topic_url?results=1");
                my $html = decode 'Windows-1252', $mech->content;
                  # It claims to be in ISO-8859-1, but it ain't!
                $html =~ m!<div \s class="board_poll">(.+?)<div \s class="poll_foot">!xs or die;
                $html = $1;
                $html =~ m!<div class="poll_head">(.+?) - Results \(\d+ votes?\)</div>! or die;
                sql 'update Topics set pollquestion = ? where topic = ?',
                    decode_entities($1), $topic;
                my $option_n = 0;
                while ($html =~ m!
                           <div \s class="table_col_auto">
                           (?: <b>)? ([^<]+) (?: </b>)?
                           </div>
                           .+?
                           <div \s class="table_col_10">
                           (\d+)
                           </div>!gxs)
                   {my ($option_text, $votes) = (decode_entities($1), $2);
                    sql 'insert or replace into PollOptions values (??)',
                        $topic, ++$option_n, $option_text, $votes;}
                note "Got poll options for $topic.";}
            $time < $last_checked and next;
            # Now let's inspect the topic, but skip over any pages
            # we've already completely archived.
            my ($posts_saved) = flatsql
                'select count(*) from Posts where topic = ?',
                $topic;
            $mech->get("$topic_url?page=" .
                int($posts_saved / $o{posts_per_page}));
            my $post = $posts_saved - $posts_saved % $o{posts_per_page};
              # A counter to keep track of which post we're looking at.
              # We index posts from 1.
            for (;;)
               {my $html = decode 'Windows-1252', $mech->content;
                  # It claims to be in ISO-8859-1, but it ain't!
                while ($html =~ m!
                         <div \s class="msg_stats">
                         <a \s href="([^"]+)" \s class="name">
                         ([^<]+)
                         </a> [^|<]+ \| \s
                         <a \s name="\d+"> </a>
                         Posted \s (\S+ \s \S+ &nbsp; \S+)
                         \s \| \s
                         <a \s href="[^"]+"> message \s detail </a>
                         \s \| \s <a [^>]+ > quote </a>
                         (\s \| \s \(edited\))?
                         (?:<[^>]+>)+
                         <div \s class="msg_body">
                         (?: <div \s class="board_poll">
                             (?:.|\n)+?
                             </form> \s+ </div> \s+)?
                         (.+?)
                         </div></td></tr> !gx)
                   {++$post > $posts_saved or next;
                      # We never overwrite saved posts with new
                      # versions, since the only way a GameFAQs post
                      # can change is by being deleted.
                    my ($author_url, $author_name, $time, $edited, $message) =
                        ($1, decode_entities($2),
                         read_gamefaqs_timestamp($3),
                         ($4 ? 1 : 0), digest_post($5));
                    my ($author_id) = flatsql
                        'select userid from Users where username = ?',
                        $author_name;
                    if (!defined $author_id)
                       {$mech->get($author_url);
                        $mech->content() =~ m!<td><b>User ID</b></td><td>(\d+)</td>! or die;
                        $author_id = $1;
                        sql 'insert into Users values (??)',
                            $author_id, $author_name;
                        $mech->back;}
                    # print "alive 5\n";
                    sql 'insert into Posts values (??)',
                        $topic, $post, $author_id, $time, $message, $edited;}
                $mech->get
                   ($mech->find_link(text => 'Next Page') or
                    $mech->find_link(text => 'Last Page') or
                    last);}
            $post > $posts_saved or croak sprintf
                'Failed to get new posts (#%03d and up) in topic %d',
                $posts_saved + 1, $topic;
            note "Saved topic $topic.";}
        $next_board_page or last;
        $mech->get($next_board_page);
        $topic_list = $mech->content;}
    
    sql q(update Misc set mval = ? where mkey = 'last_checked'),
        $checktime;
    $db->commit;
    
    note 'Finished.';

    # $db is implicitly disconnected.

    return 1;}

1;
