package WWW::GameFAQs::Archive::TopicList;

=head1 COPYRIGHT

Copyright 2010 Kodi Arfer.

This program is free software. You may copy or redistribute it under the same terms as Perl itself.

=cut

use DBIx::Simple;
use DateTime;
use WWW::GameFAQs::Archive::Internal qw(ceiling htmlsafe time_zone);

use parent 'Exporter';
@WWW::GameFAQs::Archive::TopicList::EXPORT_OK =
    qw(topic_list);

my $long_interval = 14 * 24 * 60 * 60; # Two weeks of seconds

# urlfmt: a sprintf format taking one argument that produces
# a properly escaped URL. It *doesn't* need to be escaped
# for use in HTML; I'll take care of that.

sub topic_list
#
# ~~~Mandatory arguments
# database => filepath
# topic_list => urlfmt
# message_list => urlfmt
# topics_per_page => int
#
# ~~~Optional arguments
# page => int
#   This argument will be validated.
#
# ~~~Returns
# (Success or failure [boolean], HTML topic list)
#
   {my %o = (page => 0, @_);
    my $time_zone = time_zone;

    my @error = (0, sprintf "<div class='%s'>%s</div>\n",
        'wgfa-error wgfa-no-such-topic-list-page',
        '<span><strong>Error:</strong> Bad page.</span>');
    $o{page} = $o{page} =~ /\A(\d+)\z/
      ? $1
      : return @error;

    # Read from the database.

    my ($pages, @rows);
       {my $db = DBIx::Simple->connect("dbi:SQLite:dbname=$o{database}")
            or die DBIx::Simple->error;
        $db->{sqlite_unicode} = 1;
       ($pages) =
            ($db->query('select count(*) from Topics')
                or die DBIx::Simple->error)
              ->flat;
        $pages = ceiling $pages / $o{topics_per_page};
        my $query = 'select Topics.topic, title, closed, pollquestion not null, username,
                    p2.post, p2.time
              from Posts as p1,
                   Posts as p2,
                   Topics,
                   Users
              where p1.topic = Topics.topic
                and p2.topic = Topics.topic
                and p1.post = 1
                and p1.author = userid
                and p2.post = (select max(post)
                  from Posts
                  where topic = p1.topic)
              order by p2.time desc
              limit ? offset ?';
        @rows = ($db->query($query,
                $o{topics_per_page},
                $o{topics_per_page} * $o{page})
              or die DBIx::Simple->error)
          ->arrays;};
    @rows or return @error;

    # Construct the page.

    my $nav = '<div class="wgfa-topic-list-navigation">';
    if ($pages == 1)
       {$nav .= 'Only Page';}
    else
       {my $link = sub
           {my ($rel, $page, $name) = @_;
            sprintf '<a rel="%s" href="%s">%s</a>',
                $rel,
                htmlsafe(sprintf $o{topic_list}, $page),
                $name;};
        if ($o{page} == 0)
           {$nav .= "First Page of $pages";}
        else
           {$nav .= sprintf '%s • %s',
                $link->('start', 0, '⇇ First'),
                $link->('prev', $o{page} - 1, '← Previous');}
        if ($o{page} > 0 and $o{page} < $pages - 1)
           {$nav .= ' • Page ' . ($o{page} + 1) . " of $pages • ";}
        else
           {$nav .= ' • ';}
        if ($o{page} == $pages - 1)
           {$nav .= "Last Page of $pages";}
        else
           {$nav .= sprintf '%s • %s',
                $link->('next', $o{page} + 1, 'Next →'),
                $link->('last', $pages - 1, 'Last ⇉');}}
    $nav .= "</div>\n";

    my $s .= "$nav<table class='wgfa-topic-list'>\n" .
        "<tr><th>Last Post<th>Topic<th>Author<th>Ps</tr>\n";
    my $prev_last_post;
    foreach (@rows)
       {my ($topic, $title, $closed, $is_poll, $author, $posts, $last_post) = @$_;
        $s .= sprintf "<tr class='%s'><td %s>%s<td>%s<td>%s<td %s>%s</tr>\n",
            'wgfa-topic-row' .
                ($closed ? ' wgfa-topic-closed' : '') .
                (defined($prev_last_post) &&
                 $prev_last_post - $last_post >= $long_interval
                  ? ' wgfa-topic-after-long-interval' : ''),
            ' class="wgfa-topic-last-post"',
            sprintf('<time class="wgfa-date" datetime="%s">%s</time>',
                DateTime->from_epoch(epoch => $last_post)->iso8601() . 'Z',
                DateTime->
                    from_epoch
                       (epoch => $last_post,
                        time_zone => $time_zone)->
                    format_cldr('d MMM y')),
            sprintf('%s<a href="%s">%s</a>',
                !$is_poll ? '' :
                    '<span class="wgfa-poll-marker">[Poll]</span> ',
                htmlsafe(sprintf($o{message_list}, $topic)),
                htmlsafe($title)),
            $author,
            'class="wgfa-topic-postcount"',
            $posts;
        $prev_last_post = $last_post;}
    $s .= "</table>\n$nav\n";

    return (1, $s);}

1;
